""" Frosbite code generator

$Id: freezer.py,v 1.5 2005/11/30 15:48:00 tseaver Exp $
"""
import os
import re

from Acquisition import aq_inner
from Acquisition import aq_parent
from AccessControl.SecurityInfo import ClassSecurityInfo
from Globals import InitializeClass
from Globals import package_home
from OFS.SimpleItem import SimpleItem
from Products.CMFCore.FSPageTemplate import FSPageTemplate
from Products.CMFCore.FSPythonScript import FSPythonScript
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
from Products.PythonScripts.PythonScript import PythonScript

from zope.interface import implements

from interfaces import ISkinFreezer

_VALID_PYTHON_IDENTIFIER = re.compile(r'^[_a-zA-Z][_a-zA-Z0-9]*$')
_wwwdir = os.path.join(package_home(globals()), 'www')

def _freezePythonScript(obj):
    args = ('self',)
    info = {'method_name' : obj.getId(),
            'title' : obj.title_or_id(),
           }
    lines = obj.read().splitlines()

    params = [x for x in lines if x.startswith('##parameters=')][0]
    params = params[len('##parameters='):].rstrip()
    if params:
        args += (params,)
    info['method_args'] = ', '.join(args)

    body = [x for x in lines if not x.startswith('##')]
    if not body[0].rstrip():
        body = body[1:]

    prolog = _PYTHON_METHOD_PROLOG % info
    epilog = _PYTHON_METHOD_EPILOG % info
    return {'method_text': (prolog
                           + '\n'.join(['        %s' % x for x in body])
                           + '\n'
                           + epilog
                           + '\n'
                           )}

def _freezeFSPythonScript(obj):
    return _freezePythonScript(obj._createZODBClone())

def _freezeZopePageTemplate(obj):
    info = {'method_name' : obj.getId(),
            'template_id' : '_%s_TEMPLATE' % obj.getId().upper(),
            'template_text' : obj.read(),
           }
    _PAGE_TEMPLATE_DECL % info
    _PAGE_TEMPLATE_TEXT % info
    return {'method_text': _PAGE_TEMPLATE_DECL % info,
            'module_scope_text' : _PAGE_TEMPLATE_TEXT % info,
           }

def _freezeFSPageTemplate(obj):
    return _freezeZopePageTemplate(obj._createZODBClone())

FREEZERS = {PythonScript.meta_type: _freezePythonScript,
            FSPythonScript.meta_type: _freezeFSPythonScript,
            ZopePageTemplate.meta_type: _freezeZopePageTemplate,
            FSPageTemplate.meta_type: _freezeFSPageTemplate,
           }

class SkinFreezer(SimpleItem):
    implements(ISkinFreezer)
    security = ClassSecurityInfo()
    meta_type = 'CMF Skin Freezer'

    manage_options = (({'label' : 'Freeze',
                        'action' : 'manage_freeze',
                       },
                      )
                     + SimpleItem.manage_options
                     )

    manage_freeze = PageTemplateFile('freeze.zpt', _wwwdir)

    def listCandidates(self, skins_tool=None, skin_name=None):
        """ See ISkinFreezer.
        """
        skins_tool = self._findTool(skins_tool)

        if skin_name is None:  # use tool's default skin
            skin_name = skins_tool.getDefaultSkin()

        path = skins_tool.getSkinPath(skin_name)

        return [self._getCandidateInfo(x)
                    for x in self._findCandidates(skins_tool, path)]

    def freezeCandidate(self, path, skins_tool=None):
        """ See ISkinFreezer.
        """
        skins_tool = self._findTool(skins_tool)
        obj = skins_tool.unrestrictedTraverse(path)
        return FREEZERS[obj.meta_type](obj)

    def freezeCandidates(self, product_name, class_name, paths,
                         skins_tool=None):
        """ See ISkinFreezer.
        """
        module_scope_texts = []
        method_texts = []
        skins_tool = self._findTool(skins_tool)
        info = {'product_name': product_name,
                'class_name' : class_name,
                'skins_tool_url' : skins_tool.absolute_url(),
                'paths' : '\n'.join( [' - %s' % x for x in paths]),
               }

        all_texts = [_PRODUCT_PROLOG % info ]
        for path in paths:
            frozen = self.freezeCandidate(path, skins_tool)
            module_scope_texts.append(frozen.get('module_scope_text'))
            method_texts.append(frozen['method_text'])

        all_texts.extend(module_scope_texts)
        all_texts.append(_CLASS_PROLOG % info)
        all_texts.extend(method_texts)
        all_texts.append(_CLASS_EPILOG % info)
        all_texts.append(_PRODUCT_EPILOG % info)

        return '\n'.join(filter(None, all_texts))

    def manage_freezeCandidates(self, product_name, class_name, paths,
                                skins_tool, inline, RESPONSE):
        """ Generate the Python module for download.
        """
        if not inline:
            RESPONSE.setHeader('Content-type', 'text/plain')
            RESPONSE.setHeader('Content-disposition',
                               'attachment; filename=__init__.py')
            RESPONSE.setBody(self.freezeCandidates(product_name, class_name,
                                                   paths, skins_tool))
        else:
            RESPONSE.setBody('<pre>%s</pre>' 
                                % self.freezeCandidates(product_name,
                                                        class_name,
                                                        paths,
                                                        skins_tool))

    def _findTool(self, skins_tool):
        """
        """
        if skins_tool is None:
            skins_tool = 'portal_skins'

        if isinstance(skins_tool, basestring):
            skins_tool = getattr(aq_parent(aq_inner(self)), skins_tool)

        return skins_tool

    def _findCandidates(self, skins_tool, path):
        """
        """
        meta_types = FREEZERS.keys()
        results = skins_tool.objectItems(meta_types)
        results.sort()
        seen = dict([(x[0], True) for x in results])

        if path is not None:

            for element in list(path.split(",")):

                subobj = skins_tool.unrestrictedTraverse(element)

                for path, obj in skins_tool.ZopeFind(obj=subobj,
                                                     obj_metatypes=meta_types,
                                                     search_sub=True,
                                                    ):
                    
                    obj_id = obj.getId()

                    if not _VALID_PYTHON_IDENTIFIER.match(obj_id):
                        continue

                    if obj_id not in seen:
                        seen[obj.getId()] = True
                        results.append(('%s/%s' % (element, path), obj))

        return results

    def _getCandidateInfo(self, candidate):
        """
        """
        path, obj = candidate
        return {'id' : obj.getId(),
                'path' : path,
                'url' : obj.absolute_url(),
                'object' : obj,
                'product' : None, # XXX
                'meta_type' : obj.meta_type,
                'title' : obj.title_or_id(),
               }

InitializeClass(SkinFreezer)

_PRODUCT_PROLOG = '''\
""" %(product_name)s product initialization.

Generated by Frostbite from:

 Skins tool: %(skins_tool_url)s

 Paths:
%(paths)s

$Id: freezer.py,v 1.5 2005/11/30 15:48:00 tseaver Exp $
"""
from AccessControl.SecurityInfo import ClassSecurityInfo
from Acquisition import aq_inner
from Acquisition import aq_parent
from Globals import InitializeClass
from OFS.SimpleItem import SimpleItem
from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
from Products.CMFCore.permissions import ManagePortal
from Products.CMFCore.permissions import View
from RestrictedPython.Utilities import same_type
from Shared.DC.Scripts.Script import Script
from Shared.DC.Scripts.Signature import _setFuncSignature

class AcquireableMethod(Script):
    """ Wrapper to allow methods to be acquired from freezer.
    """
    __need_name__ = 1
    index_html = None
    func_defaults=()
    func_code=None

    def __init__(self, method, id=None):
        self._m = method
        if id is None:
            id = method.__name__
        self._id = id
        defaults = getattr(method, 'func_defaults', None)
        func_code = getattr(method, 'func_code', None)
        varnames = getattr(func_code, 'co_varnames', ())
        argcount = getattr(func_code, 'co_argcount', -1)
        if varnames and varnames[0] == 'self':
            varnames = tuple(varnames[1:])
            argcount = argcount - 1
        _setFuncSignature(self,
                          defaults=defaults,
                          varnames=varnames,
                          argcount=argcount
                         )

    def __call__(self, *args, **kw):
        return self._m(aq_parent(self), *args, **kw)

    def _exec(self, bound_names, args, kw):
        """Call a Python Script

        Calling a Python Script is an actual function invocation.
        """
        return self._m(aq_parent(self), *args, **kw)

_OVERVIEW_TEMPLATE = """\\
<dtml-var manage_page_header>
<dtml-var manage_tabs>

<h3> Freezer: Speedy </h3>

<pre>
""" + __doc__ + """
</pre>

<dtml-var manage_page_footer>
"""

'''

_PAGE_TEMPLATE_TEXT = '''\
%(template_id)s = """\\
%(template_text)s
"""
'''

_CLASS_PROLOG = '''\

class %(class_name)s(SimpleItem):
    """ Frostbite-generated skin replacement
    """
    meta_type = 'Frostbite: %(class_name)s'
    security = ClassSecurityInfo()

    def manage_overview(self, REQUEST):
        """ Provide an overview of the freezer.
        """
        from DocumentTemplate import HTML
        return HTML(_OVERVIEW_TEMPLATE)(self, REQUEST)

    manage_options = (({'label' : 'Overview', 'action' : 'manage_overview'},)
                     + SimpleItem.manage_options
                     )

'''

_PYTHON_METHOD_PROLOG = '''\
    security.declareProtected(View, '%(method_name)s')
    def %(method_name)s(%(method_args)s):
        """ %(title)s
        """
        container = context = script = self
'''

_PYTHON_METHOD_EPILOG = '''\
    %(method_name)s = AcquireableMethod(%(method_name)s)
'''

_PAGE_TEMPLATE_DECL = '''\
    security.declareProtected(View, '%(method_name)s')
    %(method_name)s = ZopePageTemplate('%(method_name)s', %(template_id)s)
'''

_CLASS_EPILOG = '''\

InitializeClass(%(class_name)s)
'''

_PRODUCT_EPILOG = '''\

_FACTORY_FORM_TEMPLATE = """\\
<h1 tal:replace="structure context/manage_page_header">HEADER</h1>

<h3> Add Freezer: %(class_name)s </h3>

<form action="add%(class_name)s" method="POST">
<table>

 <tr>
  <th> Freezer ID: </th>
  <td>
   <input type="text" name="id" value="" />
  </td>
 </tr>

 <tr>
  <td />
  <td>
   <input type="submit" value=" Create Freezer " />
  </td>
 </tr>

</table>
</form>

<h1 tal:replace="structure context/manage_page_footer">FOOTER</h1>
"""

_FACTORY_FORM_ID = 'add' + '%(class_name)s' + 'Form'

add%(class_name)sForm = ZopePageTemplate(_FACTORY_FORM_ID,
                                         _FACTORY_FORM_TEMPLATE)

def add%(class_name)s(dispatcher, id, REQUEST):
    """ Add a Frostbite-generated skin replacement to dispatcher.
    """
    freezer = %(class_name)s()
    freezer._setId(id)
    dispatcher._setObject(id, freezer)
    REQUEST['RESPONSE'].redirect(dispatcher.absolute_url() + '/manage_main')

def initialize(context):
    context.registerClass(%(class_name)s,
                          constructors=(add%(class_name)sForm,
                                        add%(class_name)s,),
                          permission=ManagePortal,
                         )
'''

def addSkinFreezer(site, REQUEST=None):
    """ Add a skin freezer tool to site.
    """
    freezer = SkinFreezer()
    freezer._setId('portal_skinfreezer')
    site._setObject('portal_skinfreezer', freezer)

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect('%s/manage_workspace'
                                        % site.absolute_url())
    else:
        return site._getOb('portal_skinfreezer')
